/** @module : cell_agent
 *  @author : Adaptive & Secure Computing Systems (ASCS) Laboratory

 *  Copyright (c) 2018 ABAQS (ASCS/ECE/BU)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.

 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */

module cell_agent #(
  parameter BIT_WIDTH      = 32,
  parameter NUM_REGS       = 4,                     // Number of registers to store variables or data
  parameter NUM_BITS       = NUM_REGS * BIT_WIDTH,  // Number of bits in all data regs
  parameter TYPE_BIT_WIDTH = 1,                     // Number of bits in the cell type register
  parameter RULE_DELAY     = 1,                     // Number of clock cycles needed for valid rule output
  parameter START_TYPE     = {TYPE_BIT_WIDTH{1'b0}} // Cell type set during reset_n condition
)(
  input clock,
  input reset_n,

  // signals to get the value of space
  input [BIT_WIDTH-1:0] space_data_in, // data from local space
  input space_empty_in,                // empty signal from local space
  output space_read_rq_out,            // read the value of space

  // signals to add to the value of space
  input space_read_rq_in,
  output space_empty_out,             // empty to space
  output [BIT_WIDTH-1:0] value_added, // The amount of chemical added to local space

  // Signals to move Agent to adjacent PE
  input [NUM_BITS-1:0] cell_state_n, // north
  input [NUM_BITS-1:0] cell_state_s, // south
  input [NUM_BITS-1:0] cell_state_e, // east
  input [NUM_BITS-1:0] cell_state_w, // west

  output [NUM_BITS-1:0] cell_state_out,

  /////////////////
  // Sender Side //
  /////////////////

  // Type0 signals from neighbors
  // type 0 = empty cell agent execution unit
  input type0_n_in,
  input type0_s_in,
  input type0_e_in,
  input type0_w_in,

  // Move request signals to neighbors
  output move_rq_n_out,
  output move_rq_s_out,
  output move_rq_e_out,
  output move_rq_w_out,

  // Update signals from neighbors
  // These signals confirm that the cell movement was successful
  input type_update_n_in,
  input type_update_s_in,
  input type_update_e_in,
  input type_update_w_in,

  ///////////////////
  // Reciever Side //
  ///////////////////

  // Type0 signal to neighbors
  // type 0 = empty cell agent execution unit
  output type0_out,

  // Move request signals from neighbors
  input move_rq_n_in,
  input move_rq_s_in,
  input move_rq_e_in,
  input move_rq_w_in,

  // Update signals to neighbors
  // These signals confirm that the cell movement was successful
  output type_update_n_out,
  output type_update_s_out,
  output type_update_e_out,
  output type_update_w_out,

  // Data Shift Register
  input write_shift_reg,
  input read_shift_reg,
  input [BIT_WIDTH-1:0] shift_in,
  output [BIT_WIDTH-1:0] shift_out,

  // Type Shift Register
  input [TYPE_BIT_WIDTH-1:0] shift_t_in,
  output [TYPE_BIT_WIDTH-1:0] shift_t_out

);

wire move_rq_r;                               // move request signal from rules to control module
wire current_type;                            // type output from control reg
wire local_write_rq;                          // connect control to write_rq on local fifo
wire local_empty;                             // connect empty from local fifo to control
wire [1:0] move_direction;                    // Direction of move request from rules module
wire [BIT_WIDTH-1:0] local_value;             // The value of the field at this point in space
wire [BIT_WIDTH-1:0] rules_quantity;          // Connects the ouput of rules to control
wire [BIT_WIDTH-1:0] cell_quantity;           // Connects the ouput of control to the local fifo
wire [NUM_BITS-1:0] cell_state_ctrl_to_rules;
wire [NUM_BITS-1:0] cell_state_rules_to_ctrl;

assign space_empty_out = local_empty;
assign cell_state_out  = cell_state_ctrl_to_rules; // output the value of the rules reg

cell_control #(
  .BIT_WIDTH(BIT_WIDTH),
  .NUM_REGS(NUM_REGS),
  .NUM_BITS(NUM_BITS),
  .TYPE_BIT_WIDTH(TYPE_BIT_WIDTH),
  .RULE_DELAY(RULE_DELAY),
  .START_TYPE(START_TYPE)
)
control (
  .clock(clock),
  .reset_n(reset_n),
  .empty_from_local(local_empty),
  .empty_from_space(space_empty_in),
  .cell_state_in(cell_state_rules_to_ctrl),
  .cell_state_out(cell_state_ctrl_to_rules),
  .quantity_in(rules_quantity),              // Quantity added from rules (assumes cell is not empty)
  .quantity_out(cell_quantity),              // Quantity to actually add to space, 0 if empty
  .type_out(current_type),
  .write_rq_out(local_write_rq),
  .space_read_rq_out(space_read_rq_out),

  .cell_state_n(cell_state_n), // north
  .cell_state_s(cell_state_s), // south
  .cell_state_e(cell_state_e), // east
  .cell_state_w(cell_state_w), // west

  // Sender Side
  .type0_n_in(type0_n_in),
  .type0_s_in(type0_s_in),
  .type0_e_in(type0_e_in),
  .type0_w_in(type0_w_in),

  .move_rq_n_out(move_rq_n_out),
  .move_rq_s_out(move_rq_s_out),
  .move_rq_e_out(move_rq_e_out),
  .move_rq_w_out(move_rq_w_out),

  .type_update_n_in(type_update_n_in),
  .type_update_s_in(type_update_s_in),
  .type_update_e_in(type_update_e_in),
  .type_update_w_in(type_update_w_in),

  // From Rule Module
  .move_rq_r_in(move_rq_r),
  .move_direction(move_direction),

  // Reciever Side
  .type0_out(type0_out),

  .move_rq_n_in(move_rq_n_in),
  .move_rq_s_in(move_rq_s_in),
  .move_rq_e_in(move_rq_e_in),
  .move_rq_w_in(move_rq_w_in),

  .type_update_n_out(type_update_n_out),
  .type_update_s_out(type_update_s_out),
  .type_update_e_out(type_update_e_out),
  .type_update_w_out(type_update_w_out),

  // Data Shift Register
  .write_shift_reg(write_shift_reg),
  .read_shift_reg(read_shift_reg),
  .shift_in(shift_in),
  .shift_out(shift_out),

  // Type Shift Register
  .shift_t_in(shift_t_in),
  .shift_t_out(shift_t_out)

);


cell_rules #(
  .BIT_WIDTH(BIT_WIDTH),
  .NUM_REGS(NUM_REGS),
  .NUM_BITS(NUM_BITS),
  .TYPE_BIT_WIDTH(TYPE_BIT_WIDTH)
)
rules (
  .clock(clock),
  .reset_n(reset_n),
  .space_data_in(space_data_in),
  .cell_state_in(cell_state_ctrl_to_rules),
  .cell_state_out(cell_state_rules_to_ctrl),
  .type_in(current_type),
  .quantity(rules_quantity),                 // ammount of chemical to add to local space
                                             // assuming that the cell is not empty
  .move_rq(move_rq_r),
  .move_direction(move_direction)
);


fifo_1 #(
  .BIT_WIDTH(BIT_WIDTH)
)
cell_fifo (
  .clock(clock),
  .reset_n(reset_n),
  .data_in(cell_quantity),    // from control
  .write_rq(local_write_rq),  // from control
  .data_out(value_added),     // to space
  .read_rq(space_read_rq_in), // from space
  .empty(local_empty)         // to control and space
);

endmodule
