/** @module : cell_control
 *  @author : Adaptive & Secure Computing Systems (ASCS) Laboratory

 *  Copyright (c) 2018 ABAQS (ASCS/ECE/BU)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.

 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */

/* This module generates all of the necessary control signals for updating
 * cell states, including cell movement and reads an writes to the shift
 * register to init or save the current state.
 */

module cell_control #(
  parameter BIT_WIDTH      = 32,                         // Bit width of the data
  parameter NUM_REGS       = 4,                          // Number of registers to store variables or data
  parameter NUM_BITS       = NUM_REGS * BIT_WIDTH,       // Number of bits in all data regs
  parameter TYPE_BIT_WIDTH = 1,                          // Number of bits in the cell type register
  parameter RULE_DELAY     = 1,                          // Number of clock cycles needed for valid rule output
  parameter START_TYPE     = {TYPE_BIT_WIDTH{1'b0}},     // The type to set on system reset_n
  parameter TYPE_BITS      = TYPE_BIT_WIDTH*(NUM_REGS+1) // The number of bits in the type
                                                         // shift register. This length
                                                         // accounts for the chem field spot
                                                         // in the register so the type shift reg
                                                         // matches the data shift reg
)(
  input clock,
  input reset_n,

  input empty_from_local,                   // the empty signal from the local fifo
  input empty_from_space,                   // the empty signal from the space agent fifo

  input [NUM_BITS-1:0] cell_state_in,

  input [NUM_BITS-1:0] cell_state_n,        // north
  input [NUM_BITS-1:0] cell_state_s,        // south
  input [NUM_BITS-1:0] cell_state_e,        // east
  input [NUM_BITS-1:0] cell_state_w,        // west

  output reg [NUM_BITS-1:0] cell_state_out,

  output reg [TYPE_BIT_WIDTH-1:0] type_out,

  output reg write_rq_out,                  // To local fifo
  output reg space_read_rq_out,

  // Value of chemical concentration from/to the space agent
  input [BIT_WIDTH-1:0] quantity_in,
  output [BIT_WIDTH-1:0] quantity_out,

  /////////////////
  // Sender Side //
  /////////////////

  // Type0 signals from neighbors
  // type 0 = empty cell agent execution unit
  input type0_n_in,
  input type0_s_in,
  input type0_e_in,
  input type0_w_in,

  // Move request signals to neighbors
  output move_rq_n_out,
  output move_rq_s_out,
  output move_rq_e_out,
  output move_rq_w_out,

  // Update signals from neighbors
  // These signals confirm that the cell movement was successful
  input type_update_n_in,
  input type_update_s_in,
  input type_update_e_in,
  input type_update_w_in,

  // From Rule Module
  input move_rq_r_in,
  input [1:0] move_direction,

  ///////////////////
  // Receiver Side //
  ///////////////////

  // Type0 signal to neighbors
  // type 0 = empty cell agent execution unit
  output type0_out,

  // Move request signals from neighbors
  input move_rq_n_in,
  input move_rq_s_in,
  input move_rq_e_in,
  input move_rq_w_in,

  // Update signals to neighbors
  // These signals confirm that the cell movement was successful
  output type_update_n_out,
  output type_update_s_out,
  output type_update_e_out,
  output type_update_w_out,

  // Data Shift Register
  input write_shift_reg,
  input read_shift_reg,
  input [BIT_WIDTH-1:0] shift_in,
  output [BIT_WIDTH-1:0] shift_out,

  // Type Shift Register
  input [TYPE_BIT_WIDTH-1:0] shift_t_in,
  output [TYPE_BIT_WIDTH-1:0] shift_t_out

);


reg new_data;
reg [7:0] count;
reg [NUM_BITS-1:0] shift_reg;
reg [TYPE_BITS-1:0] shift_t_reg;

wire cell_moved;
wire [NUM_BITS-1:0] r_incoming_ns;
wire [NUM_BITS-1:0] r_incoming_ew;
wire [NUM_BITS-1:0] r_incoming;    // The cell moving into this PE

assign type0_out = type_out == {TYPE_BIT_WIDTH{1'b0}};

// Assert the move request signal for the appropriate neighbor if these
// conditions are met:
// 1. this cell agent is asserting the move request
// 2. the move_direction signal matches the output direction
// 3. the neighboring cell agent is empty
// 4. this cell agent is not empty
assign move_rq_n_out = move_rq_r_in & (move_direction == 2'h0) & type0_n_in & !type0_out;
assign move_rq_s_out = move_rq_r_in & (move_direction == 2'h1) & type0_s_in & !type0_out;
assign move_rq_e_out = move_rq_r_in & (move_direction == 2'h2) & type0_e_in & !type0_out;
assign move_rq_w_out = move_rq_r_in & (move_direction == 2'h3) & type0_w_in & !type0_out;

// Confirm move requests based on a fixed priority: north > south > east > west
// TODO: Make the priority user configurable or random each time to prevent
//       bias for movement in one direction.
assign type_update_n_out = move_rq_n_in;
assign type_update_s_out = !move_rq_n_in & move_rq_s_in;
assign type_update_e_out = !move_rq_n_in & !move_rq_s_in & move_rq_e_in;
assign type_update_w_out = !move_rq_n_in & !move_rq_s_in & !move_rq_e_in & move_rq_w_in;

assign cell_moved    = type_update_n_in | type_update_s_in |
                       type_update_e_in | type_update_w_in;

assign r_incoming_ns = move_rq_n_in ? cell_state_n : cell_state_s;
assign r_incoming_ew = move_rq_e_in ? cell_state_e : cell_state_w;
assign r_incoming    = (move_rq_n_in | move_rq_s_in) ? r_incoming_ns : r_incoming_ew;

assign quantity_out  = type0_out ? {BIT_WIDTH{1'b0}} : quantity_in;

assign shift_out     = shift_reg[BIT_WIDTH-1:0];
assign shift_t_out   = shift_t_reg[TYPE_BIT_WIDTH-1:0];

// Set the type of the local agent
always@(posedge clock or negedge reset_n) begin
  if(reset_n == 1'b0) begin
    // TODO: Change type_out value after reset_n to 0, current format is for
    // testing
    type_out <= START_TYPE;
    // TODO: add type write port & logic
  end else if( read_shift_reg ) begin
    type_out <= shift_t_reg[TYPE_BIT_WIDTH-1:0];
  end else if( cell_moved ) begin
    type_out <= {TYPE_BIT_WIDTH{1'b0}};
  end else if( move_rq_n_in | move_rq_s_in | move_rq_e_in | move_rq_w_in) begin
    // TODO: Read type from source
    type_out <= 1'b1;
  end else begin
    type_out <= type_out;
  end
end



// count number of clock ticks after inputs are all valid
always@(posedge clock or negedge reset_n) begin
  if(reset_n == 1'b0) begin
    count <= 8'd0;
  end else if( empty_from_space == 1'b1) begin
    count <= 8'd0;
  end else if( count == 8'hff) begin
    // prevent overflow in case space agent takes more than 255 cycles to
    // update
    count <= 8'hff;
  end else begin
    count <= count + 8'd1;
  end
end

// Write incoming cell to local agent reg or ...
// Write ri to ro after <RULE_DELAY> clock ticks
// TODO: Convert this block to a FSM. It could be combind with the block below
always@(posedge clock or negedge reset_n) begin
  if(reset_n == 1'b0) begin
    cell_state_out <= {NUM_BITS{1'b0}};
  end else if ( read_shift_reg ) begin
    cell_state_out <= shift_reg;
  end else if ( move_rq_n_out | move_rq_s_out | move_rq_e_out | move_rq_w_out ) begin
    cell_state_out <= r_incoming;
  end else if ( (count == RULE_DELAY) & !type0_out ) begin
    cell_state_out <= cell_state_in;
  end else begin
    cell_state_out <= cell_state_out;
  end
end

// Send read_rq signal to indicate that the data has been read
// TODO: convert this whole block to a FSM
always@(posedge clock or negedge reset_n) begin
  if(reset_n == 1'b0) begin
    space_read_rq_out <= 1'b0;
    new_data <= 1'b1; // write 0 to space after reset_n
  end else if ( move_rq_n_out | move_rq_s_out | move_rq_e_out | move_rq_w_out ) begin
    space_read_rq_out <= 1'b0;
    new_data <= 1'b0;
  end else if ( (count >= RULE_DELAY) & (empty_from_space == 1'b0) & (new_data == 1'b0) ) begin
    // TODO: This section should not have to wait for RULE delay if the cell
    // is empty. Add logic to update early in this case. This whole block
    // should be converted to a FSM to improve performance
    space_read_rq_out <= 1'b1;
    new_data <= 1'b1;
  end else begin
    space_read_rq_out <= 1'b0;
    new_data <= new_data & ~write_rq_out; // if wr == 1 then set new data to 0
  end
end


// write to space agent
always@(posedge clock or negedge reset_n) begin
  if(reset_n == 1'b0) begin
    write_rq_out <= 1'b0;
  end else begin
    write_rq_out <= empty_from_local & new_data;
  end
end


// Shift Register
always@(posedge clock or negedge reset_n) begin
  if(reset_n == 1'b0) begin
    shift_reg <= {NUM_BITS{1'b0}};
  end else if ( write_shift_reg ) begin
    shift_reg <= cell_state_out;
  end else begin
    shift_reg <= { shift_in, shift_reg[NUM_BITS-1:BIT_WIDTH] };
  end
end

// Type Shift Register
always@(posedge clock or negedge reset_n) begin
  if(reset_n == 1'b0) begin
    shift_t_reg <= {TYPE_BITS{1'b0}};
  end else if ( write_shift_reg ) begin
    shift_t_reg <= {NUM_REGS{type_out}};
  end else begin
    shift_t_reg <= { shift_t_in, shift_t_reg[TYPE_BITS-TYPE_BIT_WIDTH-1:TYPE_BIT_WIDTH] };
  end
end


endmodule
