/** @module : processing_element
 *  @author : Adaptive & Secure Computing Systems (ASCS) Laboratory

 *  Copyright (c) 2018 ABAQS (ASCS/ECE/BU)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.

 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */

/*
 * This version of a PE implements one space agent and one cell agent. Future
 * versions will parameterize the number of cell agents and space agents in
 * each PE
 *
 * Data is read from the PE with a shift register
 * First the Chemical field value is shifted out, then the cell agent
 * variables are shifted out.
 * shift_in -> [ Cell Agent Vars ] -> [ Chem Value ] -> shift_out
 */

module processing_element #(
  parameter BIT_WIDTH      = 32,
  parameter SPACE_START    = {BIT_WIDTH{1'b0}},     // Starting value of chem field
  parameter NUM_REGS       = 4,                     // Number of registers to store variables or data
  parameter NUM_BITS       = NUM_REGS * BIT_WIDTH,  // Number of bits in all data regs
  parameter TYPE_BIT_WIDTH = 1,                     // Number of bits in the cell type register
  parameter RULE_DELAY     = 1,
  parameter START_TYPE     = {TYPE_BIT_WIDTH{1'b0}}
)(
  input clock,
  input reset_n,
  input [3:0] empty_from_neighbor,
  input [3:0] read_rq_from_neighbor,
  input [BIT_WIDTH-1:0] north_in,
  input [BIT_WIDTH-1:0] south_in,
  input [BIT_WIDTH-1:0] east_in,
  input [BIT_WIDTH-1:0] west_in,
  output [3:0] read_rq_to_neighbor,
  output [3:0] empty_to_neighbor,
  output [BIT_WIDTH-1:0] north_out,
  output [BIT_WIDTH-1:0] south_out,
  output [BIT_WIDTH-1:0] east_out,
  output [BIT_WIDTH-1:0] west_out,

  // Signals to move Agent to adjacent PE
  input [NUM_BITS-1:0] cell_state_n, // north
  input [NUM_BITS-1:0] cell_state_s, // south
  input [NUM_BITS-1:0] cell_state_e, // east
  input [NUM_BITS-1:0] cell_state_w, // west

  //input type_init, // Manually set the type of agent/cell
  output [NUM_BITS-1:0] cell_state_out,

  // Sender Side
  input type0_n_in,
  input type0_s_in,
  input type0_e_in,
  input type0_w_in,

  output move_rq_n_out,
  output move_rq_s_out,
  output move_rq_e_out,
  output move_rq_w_out,

  input type_update_n_in,
  input type_update_s_in,
  input type_update_e_in,
  input type_update_w_in,

  // Reciever Side

  input move_rq_n_in,
  input move_rq_s_in,
  input move_rq_e_in,
  input move_rq_w_in,

  output type0_out,

  output type_update_n_out,
  output type_update_s_out,
  output type_update_e_out,
  output type_update_w_out,

  // Data Shift Register
  input write_shift_reg,
  input read_shift_reg,
  input [BIT_WIDTH-1:0] shift_in,
  output [BIT_WIDTH-1:0] shift_out,

  // Type Shift Register
  input [TYPE_BIT_WIDTH-1:0] shift_t_in,
  output [TYPE_BIT_WIDTH-1:0] shift_t_out

);

wire empty_from_cell;
wire read_rq_from_cell;
wire [BIT_WIDTH-1:0] data_from_cell;
wire [BIT_WIDTH-1:0] data_to_cell;
wire read_rq_to_cell;
wire empty_to_cell;
wire local_empty_cell;
wire local_write_rq_cell;
wire [BIT_WIDTH-1:0] shift_middle;

space_agent #(
  .BIT_WIDTH(BIT_WIDTH),
  .START(SPACE_START)
)
space_pe(
  .clock(clock),
  .reset_n(reset_n),
  .empty_from_neighbor(empty_from_neighbor),
  .read_rq_from_neighbor(read_rq_from_neighbor),
  .north_in(north_in),
  .south_in(south_in),
  .east_in(east_in),
  .west_in(west_in),
  .empty_from_cell(empty_from_cell),
  .read_rq_from_cell(read_rq_from_cell),
  .cell_in(data_from_cell),
  .read_rq_to_neighbor(read_rq_to_neighbor),
  .empty_to_neighbor(empty_to_neighbor),
  .north_out(north_out),
  .south_out(south_out),
  .east_out(east_out),
  .west_out(west_out),
  .cell_out(data_to_cell),
  .read_rq_to_cell(read_rq_to_cell),
  .empty_to_cell(empty_to_cell),

  // Data Shift Register
  .write_shift_reg(write_shift_reg),
  .read_shift_reg(read_shift_reg),
  .shift_in(shift_middle),
  .shift_out(shift_out)

);


cell_agent #(
  .BIT_WIDTH(BIT_WIDTH),
  .NUM_REGS(NUM_REGS),             // Number of registers to store variables or data
  .NUM_BITS(NUM_BITS),             // Number of bits in all data regs
  .TYPE_BIT_WIDTH(TYPE_BIT_WIDTH), // Number of bits in the cell type register
  .RULE_DELAY(RULE_DELAY),
  .START_TYPE(START_TYPE)
)
cell_pe(
  .clock(clock),
  .reset_n(reset_n),
  .space_data_in(data_to_cell),
  .space_empty_in(empty_to_cell),
  .space_read_rq_out(read_rq_from_cell),
  .space_empty_out(empty_from_cell),
  .space_read_rq_in(read_rq_to_cell),
  .value_added(data_from_cell),
  .cell_state_out(cell_state_out),
  .cell_state_n(cell_state_n), // north
  .cell_state_s(cell_state_s), // south
  .cell_state_e(cell_state_e), // east
  .cell_state_w(cell_state_w), // west

  // Sender Side
  .type0_n_in(type0_n_in),
  .type0_s_in(type0_s_in),
  .type0_e_in(type0_e_in),
  .type0_w_in(type0_w_in),

  .move_rq_n_out(move_rq_n_out),
  .move_rq_s_out(move_rq_s_out),
  .move_rq_e_out(move_rq_e_out),
  .move_rq_w_out(move_rq_w_out),

  .type_update_n_in(type_update_n_in),
  .type_update_s_in(type_update_s_in),
  .type_update_e_in(type_update_e_in),
  .type_update_w_in(type_update_w_in),

  // Reciever Side
  .type0_out(type0_out),

  .move_rq_n_in(move_rq_n_in),
  .move_rq_s_in(move_rq_s_in),
  .move_rq_e_in(move_rq_e_in),
  .move_rq_w_in(move_rq_w_in),

  .type_update_n_out(type_update_n_out),
  .type_update_s_out(type_update_s_out),
  .type_update_e_out(type_update_e_out),
  .type_update_w_out(type_update_w_out),

  // Data Shift Register
  .write_shift_reg(write_shift_reg),
  .read_shift_reg(read_shift_reg),
  .shift_in(shift_in),
  .shift_out(shift_middle),

  // Type Shift Register
  .shift_t_in(shift_t_in),
  .shift_t_out(shift_t_out)

);


endmodule
