/** @module : space_agent
 *  @author : Adaptive & Secure Computing Systems (ASCS) Laboratory

 *  Copyright (c) 2018 ABAQS (ASCS/ECE/BU)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.

 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */


module space_agent #(
  parameter BIT_WIDTH = 32,
  parameter START     = {BIT_WIDTH{1'b0}}
)(
  input clock,
  input reset_n,
  input [3:0] empty_from_neighbor,
  input [3:0] read_rq_from_neighbor,
  input [BIT_WIDTH-1:0] north_in,
  input [BIT_WIDTH-1:0] south_in,
  input [BIT_WIDTH-1:0] east_in,
  input [BIT_WIDTH-1:0] west_in,
  input empty_from_cell,
  input read_rq_from_cell,
  input [BIT_WIDTH-1:0] cell_in,
  output [3:0] read_rq_to_neighbor,
  output [3:0] empty_to_neighbor,
  output [BIT_WIDTH-1:0] north_out,
  output [BIT_WIDTH-1:0] south_out,
  output [BIT_WIDTH-1:0] east_out,
  output [BIT_WIDTH-1:0] west_out,
  output read_rq_to_cell,
  output empty_to_cell,
  output [BIT_WIDTH-1:0] cell_out,

  // Data Shift Register
  input write_shift_reg,
  input read_shift_reg,
  input [BIT_WIDTH-1:0] shift_in,
  output [BIT_WIDTH-1:0] shift_out

);

wire [BIT_WIDTH-1:0] local_value; // The value of the field at this point in space

wire [3:0] local_write_rq; // connect core to write_rq on local fifos
wire local_write_rq_cell;

wire [3:0] local_empty; // connect empty from local fifos to core
wire local_empty_cell;



assign empty_to_neighbor = local_empty;
assign empty_to_cell     = local_empty_cell;

space_agent_core #(
  .BIT_WIDTH(BIT_WIDTH),
  .START(START)
)
core (
  .clock(clock),
  .reset_n(reset_n),
  .empty_from_neighbor(empty_from_neighbor),
  .empty_from_cell(empty_from_cell),
  .north_in(north_in),
  .south_in(south_in),
  .east_in(east_in),
  .west_in(west_in),
  .cell_in(cell_in),
  // NSEW neighbor signals
  .empty_from_local(local_empty),
  .write_rq_out(local_write_rq),
  .read_rq_out(read_rq_to_neighbor),
  // Cell Agent signals
  .empty_from_local_cell(local_empty_cell),
  .cell_write_rq_out(local_write_rq_cell),
  .cell_read_rq_out(read_rq_to_cell),

  .data_out(local_value),

  // Data Shift Register
  .write_shift_reg(write_shift_reg),
  .read_shift_reg(read_shift_reg),
  .shift_in(shift_in),
  .shift_out(shift_out)

);

/**************
* Local Fifos *
**************/
fifo_1 #(
  .BIT_WIDTH(BIT_WIDTH)
)
north (
  .clock(clock),
  .reset_n(reset_n),
  .data_in(local_value),
  .write_rq(local_write_rq[0]),
  .data_out(north_out),
  .read_rq(read_rq_from_neighbor[0]),
  .empty(local_empty[0])
);

fifo_1 #(
  .BIT_WIDTH(BIT_WIDTH)
)
south (
  .clock(clock),
  .reset_n(reset_n),
  .data_in(local_value),
  .write_rq(local_write_rq[1]),
  .data_out(south_out),
  .read_rq(read_rq_from_neighbor[1]),
  .empty(local_empty[1])
);

fifo_1 #(
  .BIT_WIDTH(BIT_WIDTH)
)
east (
  .clock(clock),
  .reset_n(reset_n),
  .data_in(local_value),
  .write_rq(local_write_rq[2]),
  .data_out(east_out),
  .read_rq(read_rq_from_neighbor[2]),
  .empty(local_empty[2])
);

fifo_1 #(
  .BIT_WIDTH(BIT_WIDTH)
)
west(
  .clock(clock),
  .reset_n(reset_n),
  .data_in(local_value),
  .write_rq(local_write_rq[3]),
  .data_out(west_out),
  .read_rq(read_rq_from_neighbor[3]),
  .empty(local_empty[3])
);

fifo_1 #(
  .BIT_WIDTH(BIT_WIDTH)
)
cell_fifo (
  .clock(clock),
  .reset_n(reset_n),
  .data_in(local_value),
  .write_rq(local_write_rq_cell),
  .data_out(cell_out),
  .read_rq(read_rq_from_cell),
  .empty(local_empty_cell)
);

endmodule
