/** @module : torus
 *  @author : Adaptive & Secure Computing Systems (ASCS) Laboratory

 *  Copyright (c) 2018 ABAQS (ASCS/ECE/BU)
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.

 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */

module torus #(
  parameter BIT_WIDTH      = 32,                    // The bit width of the data
  parameter NUM_REGS       = 4,                     // Number of registers to store variables or data
  parameter NUM_BITS       = NUM_REGS * BIT_WIDTH,  // Number of bits in all data regs
  parameter TYPE_BIT_WIDTH = 1,                     // Number of bits in the cell type register
  parameter RULE_DELAY     = 1,                     // Number of clock cycles needed for valid rule output
  parameter WIDTH          = 3,                     // Number of processing elements in each simulation row
  parameter HEIGHT         = 3,                     // Number of processing elements in each simulation column
  parameter START_TYPE     = {TYPE_BIT_WIDTH{1'b0}} // The type to set on system reset_n
) (
  input clock,
  input reset_n,

  input write_shift_reg,
  input read_shift_reg,

  input [(HEIGHT*BIT_WIDTH)-1:0] data_vector_in,
  output [(HEIGHT*BIT_WIDTH)-1:0] data_vector_out,

  input [(HEIGHT*TYPE_BIT_WIDTH)-1:0] type_vector_in,
  output [(HEIGHT*TYPE_BIT_WIDTH)-1:0] type_vector_out

);

wire [3:0] read_rq [WIDTH-1:0][HEIGHT-1:0];
wire [3:0] empty   [WIDTH-1:0][HEIGHT-1:0];

wire [BIT_WIDTH-1:0] space_fifo_n [WIDTH-1:0][HEIGHT-1:0]; // fifo outputs from each PE
wire [BIT_WIDTH-1:0] space_fifo_s [WIDTH-1:0][HEIGHT-1:0]; // fifo outputs from each PE
wire [BIT_WIDTH-1:0] space_fifo_e [WIDTH-1:0][HEIGHT-1:0]; // fifo outputs from each PE
wire [BIT_WIDTH-1:0] space_fifo_w [WIDTH-1:0][HEIGHT-1:0]; // fifo outputs from each PE

// Values of the registers in each Cell Agent Execution Unit
wire [NUM_BITS-1:0] cell_state [WIDTH-1:0][HEIGHT-1:0];

wire move_rq_n [WIDTH-1:0][HEIGHT-1:0];
wire move_rq_s [WIDTH-1:0][HEIGHT-1:0];
wire move_rq_e [WIDTH-1:0][HEIGHT-1:0];
wire move_rq_w [WIDTH-1:0][HEIGHT-1:0];

wire type0 [WIDTH-1:0][HEIGHT-1:0];

wire update_n [WIDTH-1:0][HEIGHT-1:0];
wire update_s [WIDTH-1:0][HEIGHT-1:0];
wire update_e [WIDTH-1:0][HEIGHT-1:0];
wire update_w [WIDTH-1:0][HEIGHT-1:0];

wire [BIT_WIDTH-1:0]      shift_reg   [WIDTH-2:0][HEIGHT-1:0];
wire [TYPE_BIT_WIDTH-1:0] shift_reg_t [WIDTH-2:0][HEIGHT-1:0];

genvar i;
genvar j;
generate
// Generate interior processing elements, no sides or corners
for(i=1; i < WIDTH-1; i=i+1) begin : interior_in_loop // x loop
  for(j=1; j < HEIGHT-1; j=j+1) begin : interior_j_loop // y loop
    processing_element #(
      .BIT_WIDTH(BIT_WIDTH),
      .SPACE_START(32'h00000000),
      .NUM_REGS(NUM_REGS),
      .NUM_BITS(NUM_BITS),
      .TYPE_BIT_WIDTH(TYPE_BIT_WIDTH),
      .RULE_DELAY(RULE_DELAY),
      .START_TYPE(START_TYPE)
    )
    interior (
      .clock(clock),
      .reset_n(reset_n),
      .read_rq_from_neighbor( { read_rq[i-1][j][2], read_rq[i+1][j][3],
                                read_rq[i][j-1][0], read_rq[i][j+1][1] } ),
      .empty_from_neighbor( { empty[i-1][j][2], empty[i+1][j][3],
                              empty[i][j-1][0], empty[i][j+1][1] } ),

      .north_in(space_fifo_s[i][j+1]),
      .south_in(space_fifo_n[i][j-1]),
      .east_in(space_fifo_w[i+1][j]),
      .west_in(space_fifo_e[i-1][j]),

      .read_rq_to_neighbor(read_rq[i][j]),
      .empty_to_neighbor(empty[i][j]),

      .north_out(space_fifo_n[i][j]),
      .south_out(space_fifo_s[i][j]),
      .east_out(space_fifo_e[i][j]),
      .west_out(space_fifo_w[i][j]),

      // Signals to move Agent to adjacent PE
      .cell_state_n(cell_state[i][j+1]), // north
      .cell_state_s(cell_state[i][j-1]), // south
      .cell_state_e(cell_state[i+1][j]), // east
      .cell_state_w(cell_state[i-1][j]), // west
      .cell_state_out(cell_state[i][j]), // output

      // Sender Side
      .type0_n_in(type0[i][j+1]),
      .type0_s_in(type0[i][j-1]),
      .type0_e_in(type0[i+1][j]),
      .type0_w_in(type0[i-1][j]),

      .move_rq_n_out(move_rq_n[i][j]),
      .move_rq_s_out(move_rq_s[i][j]),
      .move_rq_e_out(move_rq_e[i][j]),
      .move_rq_w_out(move_rq_w[i][j]),

      .type_update_n_in(update_s[i][j+1]), // The south wire of the northern PE
      .type_update_s_in(update_n[i][j-1]), // The north wire of the southern PE
      .type_update_e_in(update_w[i+1][j]), // The west wire of the eastern PE
      .type_update_w_in(update_e[i-1][j]), // The east wire of the western PE

      // Reciever Side
      .type0_out(type0[i][j]),

      .move_rq_n_in(move_rq_s[i][j+1]), // The south wire of the northern PE
      .move_rq_s_in(move_rq_n[i][j-1]), // The north wire of the southern PE
      .move_rq_e_in(move_rq_w[i+1][j]), // The west wire of the eastern PE
      .move_rq_w_in(move_rq_e[i-1][j]), // The east wire of the western PE

      .type_update_n_out(update_n[i][j]),
      .type_update_s_out(update_s[i][j]),
      .type_update_e_out(update_e[i][j]),
      .type_update_w_out(update_w[i][j]),

      // Data Shift Register
      .write_shift_reg(write_shift_reg),
      .read_shift_reg(read_shift_reg),
      .shift_in(shift_reg[i-1][j]),
      .shift_out(shift_reg[i][j]),

      // Type Shift Register
      .shift_t_in(shift_reg_t[i-1][j]),
      .shift_t_out(shift_reg_t[i][j])

    );
  end // y loop
end // x loop
endgenerate

generate
// Generate top and botom side processing elements (Width dimension), no corners
for(i=1; i < WIDTH-1; i=i+1) begin : sides_in_loop // sides x loop

  processing_element #(
    .BIT_WIDTH(BIT_WIDTH),
    .SPACE_START(32'h00000000),
    .NUM_REGS(NUM_REGS),
    .NUM_BITS(NUM_BITS),
    .TYPE_BIT_WIDTH(TYPE_BIT_WIDTH),
    .RULE_DELAY(RULE_DELAY),
    .START_TYPE(START_TYPE)
  )
  top_side (
    .clock(clock),
    .reset_n(reset_n),
    .read_rq_from_neighbor( { read_rq[i-1][HEIGHT-1][2], read_rq[i+1][HEIGHT-1][3],
                            read_rq[i][HEIGHT-2][0], read_rq[i][0][1] } ),
    .empty_from_neighbor( { empty[i-1][HEIGHT-1][2], empty[i+1][HEIGHT-1][3],
                            empty[i][HEIGHT-2][0], empty[i][0][1] } ),

    .north_in(space_fifo_s[i][0]),
    .south_in(space_fifo_n[i][HEIGHT-2]),
    .east_in(space_fifo_w[i+1][HEIGHT-1]),
    .west_in(space_fifo_e[i-1][HEIGHT-1]),

    .read_rq_to_neighbor(read_rq[i][HEIGHT-1]),
    .empty_to_neighbor(empty[i][HEIGHT-1]),

    .north_out(space_fifo_n[i][HEIGHT-1]),
    .south_out(space_fifo_s[i][HEIGHT-1]),
    .east_out(space_fifo_e[i][HEIGHT-1]),
    .west_out(space_fifo_w[i][HEIGHT-1]),

    // Signals to move Agent to adjacent PE
    .cell_state_n(cell_state[i][0]),          // north
    .cell_state_s(cell_state[i][HEIGHT-2]),   // south
    .cell_state_e(cell_state[i+1][HEIGHT-1]), // east
    .cell_state_w(cell_state[i-1][HEIGHT-1]), // west
    .cell_state_out(cell_state[i][HEIGHT-1]), // output

    // Sender Side
    .type0_n_in(type0[i][0]),
    .type0_s_in(type0[i][HEIGHT-2]),
    .type0_e_in(type0[i+1][HEIGHT-1]),
    .type0_w_in(type0[i-1][HEIGHT-1]),

    .move_rq_n_out(move_rq_n[i][HEIGHT-1]),
    .move_rq_s_out(move_rq_s[i][HEIGHT-1]),
    .move_rq_e_out(move_rq_e[i][HEIGHT-1]),
    .move_rq_w_out(move_rq_w[i][HEIGHT-1]),

    .type_update_n_in(update_s[i][0]),          // The south wire of the northern PE
    .type_update_s_in(update_n[i][HEIGHT-2]),   // The north wire of the southern PE
    .type_update_e_in(update_w[i+1][HEIGHT-1]), // The west wire of the eastern PE
    .type_update_w_in(update_e[i-1][HEIGHT-1]), // The east wire of the western PE

    // Reciever Side
    .type0_out(type0[i][HEIGHT-1]),

    .move_rq_n_in(move_rq_s[i][0]),          // The south wire of the northern PE
    .move_rq_s_in(move_rq_n[i][HEIGHT-2]),   // The north wire of the southern PE
    .move_rq_e_in(move_rq_w[i+1][HEIGHT-1]), // The west wire of the eastern PE
    .move_rq_w_in(move_rq_e[i-1][HEIGHT-1]), // The east wire of the western PE

    .type_update_n_out(update_n[i][HEIGHT-1]),
    .type_update_s_out(update_s[i][HEIGHT-1]),
    .type_update_e_out(update_e[i][HEIGHT-1]),
    .type_update_w_out(update_w[i][HEIGHT-1]),

    // Data Shift Register
    .write_shift_reg(write_shift_reg),
    .read_shift_reg(read_shift_reg),
    .shift_in(shift_reg[i-1][HEIGHT-1]),
    .shift_out(shift_reg[i][HEIGHT-1]),

    // Type Shift Register
    .shift_t_in(shift_reg_t[i-1][HEIGHT-1]),
    .shift_t_out(shift_reg_t[i][HEIGHT-1])

  );
  processing_element #(
    .BIT_WIDTH(BIT_WIDTH),
    .SPACE_START(32'h00000000),
    .NUM_REGS(NUM_REGS),
    .NUM_BITS(NUM_BITS),
    .TYPE_BIT_WIDTH(TYPE_BIT_WIDTH),
    .RULE_DELAY(RULE_DELAY),
    .START_TYPE(START_TYPE)
  )
  bottom_side (
    .clock(clock),
    .reset_n(reset_n),
    .read_rq_from_neighbor( { read_rq[i-1][0][2], read_rq[i+1][0][3],
                            read_rq[i][HEIGHT-1][0], read_rq[i][1][1] } ),
    .empty_from_neighbor( { empty[i-1][0][2], empty[i+1][0][3],
                            empty[i][HEIGHT-1][0], empty[i][1][1] } ),

    .north_in(space_fifo_s[i][1]),
    .south_in(space_fifo_n[i][HEIGHT-1]),
    .east_in(space_fifo_w[i+1][0]),
    .west_in(space_fifo_e[i-1][0]),

    .read_rq_to_neighbor(read_rq[i][0]),
    .empty_to_neighbor(empty[i][0]),

    .north_out(space_fifo_n[i][0]),
    .south_out(space_fifo_s[i][0]),
    .east_out(space_fifo_e[i][0]),
    .west_out(space_fifo_w[i][0]),

    // Signals to move Agent to adjacent PE
    .cell_state_n(cell_state[i][1]),        // north
    .cell_state_s(cell_state[i][HEIGHT-1]), // south
    .cell_state_e(cell_state[i+1][0]),      // east
    .cell_state_w(cell_state[i-1][0]),      // west
    .cell_state_out(cell_state[i][0]),      // output

    // Sender Side
    .type0_n_in(type0[i][1]),
    .type0_s_in(type0[i][HEIGHT-1]),
    .type0_e_in(type0[i+1][0]),
    .type0_w_in(type0[i-1][0]),

    .move_rq_n_out(move_rq_n[i][0]),
    .move_rq_s_out(move_rq_s[i][0]),
    .move_rq_e_out(move_rq_e[i][0]),
    .move_rq_w_out(move_rq_w[i][0]),

    .type_update_n_in(update_s[i][1]),        // The south wire of the northern PE
    .type_update_s_in(update_n[i][HEIGHT-1]), // The north wire of the southern PE
    .type_update_e_in(update_w[i+1][0]),      // The west wire of the eastern PE
    .type_update_w_in(update_e[i-1][0]),      // The east wire of the western PE

    // Reciever Side
    .type0_out(type0[i][0]),

    .move_rq_n_in(move_rq_s[i][1]),        // The south wire of the northern PE
    .move_rq_s_in(move_rq_n[i][HEIGHT-1]), // The north wire of the southern PE
    .move_rq_e_in(move_rq_w[i+1][0]),      // The west wire of the eastern PE
    .move_rq_w_in(move_rq_e[i-1][0]),      // The east wire of the western PE

    .type_update_n_out(update_n[i][0]),
    .type_update_s_out(update_s[i][0]),
    .type_update_e_out(update_e[i][0]),
    .type_update_w_out(update_w[i][0]),

    // Data Shift Register
    .write_shift_reg(write_shift_reg),
    .read_shift_reg(read_shift_reg),
    .shift_in(shift_reg[i-1][0]),
    .shift_out(shift_reg[i][0]),

    // Type Shift Register
    .shift_t_in(shift_reg_t[i-1][0]),
    .shift_t_out(shift_reg_t[i][0])

  );
end // x loop
endgenerate

generate
// Generate right and left side processing elements (Width dimension), no corners
for(j=1; j < HEIGHT-1; j=j+1) begin : sides_j_loop // y loop
  processing_element #(
    .BIT_WIDTH(BIT_WIDTH),
    .SPACE_START(32'h00000000),
    .NUM_REGS(NUM_REGS),
    .NUM_BITS(NUM_BITS),
    .TYPE_BIT_WIDTH(TYPE_BIT_WIDTH),
    .RULE_DELAY(RULE_DELAY),
    .START_TYPE(START_TYPE)
  )
  right_side (
    .clock(clock),
    .reset_n(reset_n),
    .read_rq_from_neighbor( { read_rq[WIDTH-2][j][2], read_rq[0][j][3],
                            read_rq[WIDTH-1][j-1][0], read_rq[WIDTH-1][j+1][1] } ),
    .empty_from_neighbor( { empty[WIDTH-2][j][2], empty[0][j][3],
                            empty[WIDTH-1][j-1][0], empty[WIDTH-1][j+1][1] } ),

    .north_in(space_fifo_s[WIDTH-1][j+1]),
    .south_in(space_fifo_n[WIDTH-1][j-1]),
    .east_in(space_fifo_w[0][j]),
    .west_in(space_fifo_e[WIDTH-2][j]),

    .read_rq_to_neighbor(read_rq[WIDTH-1][j]),
    .empty_to_neighbor(empty[WIDTH-1][j]),

    .north_out(space_fifo_n[WIDTH-1][j]),
    .south_out(space_fifo_s[WIDTH-1][j]),
    .east_out(space_fifo_e[WIDTH-1][j]),
    .west_out(space_fifo_w[WIDTH-1][j]),

    // Signals to move Agent to adjacent PE
    .cell_state_n(cell_state[WIDTH-1][j+1]), // north
    .cell_state_s(cell_state[WIDTH-1][j-1]), // south
    .cell_state_e(cell_state[0][j]),         // east
    .cell_state_w(cell_state[WIDTH-1][j]),   // west
    .cell_state_out(cell_state[WIDTH-1][j]), // output

    // Sender Side
    .type0_n_in(type0[WIDTH-1][j+1]),
    .type0_s_in(type0[WIDTH-1][j-1]),
    .type0_e_in(type0[0][j]),
    .type0_w_in(type0[WIDTH-2][j]),

    .move_rq_n_out(move_rq_n[WIDTH-1][j]),
    .move_rq_s_out(move_rq_s[WIDTH-1][j]),
    .move_rq_e_out(move_rq_e[WIDTH-1][j]),
    .move_rq_w_out(move_rq_w[WIDTH-1][j]),

    .type_update_n_in(update_s[WIDTH-1][j+1]), // The south wire of the northern PE
    .type_update_s_in(update_n[WIDTH-1][j-1]), // The north wire of the southern PE
    .type_update_e_in(update_w[0][j]),         // The west wire of the eastern PE
    .type_update_w_in(update_e[WIDTH-2][j]),   // The east wire of the western PE

    // Reciever Side
    .type0_out(type0[WIDTH-1][j]),

    .move_rq_n_in(move_rq_s[WIDTH-1][j+1]), // The south wire of the northern PE
    .move_rq_s_in(move_rq_n[WIDTH-1][j-1]), // The north wire of the southern PE
    .move_rq_e_in(move_rq_w[0][j]),         // The west wire of the eastern PE
    .move_rq_w_in(move_rq_e[WIDTH-2][j]),   // The east wire of the western PE

    .type_update_n_out(update_n[WIDTH-1][j]),
    .type_update_s_out(update_s[WIDTH-1][j]),
    .type_update_e_out(update_e[WIDTH-1][j]),
    .type_update_w_out(update_w[WIDTH-1][j]),

    // Data Shift Register
    .write_shift_reg(write_shift_reg),
    .read_shift_reg(read_shift_reg),
    .shift_in(shift_reg[WIDTH-2][j]),
    .shift_out(data_vector_out[(BIT_WIDTH*(j+1))-1:BIT_WIDTH*j]),

    // Type Shift Register
    .shift_t_in(shift_reg_t[WIDTH-2][j]),
    .shift_t_out(type_vector_out[(TYPE_BIT_WIDTH*(j+1))-1:TYPE_BIT_WIDTH*j])

  );

  processing_element #(
    .BIT_WIDTH(BIT_WIDTH),
    .SPACE_START(32'h00000000),
    .NUM_REGS(NUM_REGS),
    .NUM_BITS(NUM_BITS),
    .TYPE_BIT_WIDTH(TYPE_BIT_WIDTH),
    .RULE_DELAY(RULE_DELAY),
    .START_TYPE(START_TYPE)
  )
  left_side (
    .clock(clock),
    .reset_n(reset_n),
    .read_rq_from_neighbor( { read_rq[WIDTH-1][j][2], read_rq[1][j][3],
                            read_rq[0][j-1][0], read_rq[0][j+1][1] } ),
    .empty_from_neighbor( { empty[WIDTH-1][j][2], empty[1][j][3],
                            empty[0][j-1][0], empty[0][j+1][1] } ),

    .north_in(space_fifo_s[0][j+1]),
    .south_in(space_fifo_n[0][j-1]),
    .east_in(space_fifo_w[1][j]),
    .west_in(space_fifo_e[WIDTH-1][j]),

    .read_rq_to_neighbor(read_rq[0][j]),
    .empty_to_neighbor(empty[0][j]),

    .north_out(space_fifo_n[0][j]),
    .south_out(space_fifo_s[0][j]),
    .east_out(space_fifo_e[0][j]),
    .west_out(space_fifo_w[0][j]),

    // Signals to move Agent to adjacent PE
    .cell_state_n(cell_state[0][j+1]),     // north
    .cell_state_s(cell_state[0][j-1]),     // south
    .cell_state_e(cell_state[1][j]),       // east
    .cell_state_w(cell_state[WIDTH-1][j]), // west
    .cell_state_out(cell_state[0][j]),     // output

    // Sender Side
    .type0_n_in(type0[0][j+1]),
    .type0_s_in(type0[0][j-1]),
    .type0_e_in(type0[1][j]),
    .type0_w_in(type0[WIDTH-1][j]),

    .move_rq_n_out(move_rq_n[0][j]),
    .move_rq_s_out(move_rq_s[0][j]),
    .move_rq_e_out(move_rq_e[0][j]),
    .move_rq_w_out(move_rq_w[0][j]),

    .type_update_n_in(update_s[0][j+1]),     // The south wire of the northern PE
    .type_update_s_in(update_n[0][j-1]),     // The north wire of the southern PE
    .type_update_e_in(update_w[1][j]),       // The west wire of the eastern PE
    .type_update_w_in(update_e[WIDTH-1][j]), // The east wire of the western PE

    // Reciever Side
    .type0_out(type0[0][j]),

    .move_rq_n_in(move_rq_s[0][j+1]),     // The south wire of the northern PE
    .move_rq_s_in(move_rq_n[0][j-1]),     // The north wire of the southern PE
    .move_rq_e_in(move_rq_w[1][j]),       // The west wire of the eastern PE
    .move_rq_w_in(move_rq_e[WIDTH-1][j]), // The east wire of the western PE

    .type_update_n_out(update_n[0][j]),
    .type_update_s_out(update_s[0][j]),
    .type_update_e_out(update_e[0][j]),
    .type_update_w_out(update_w[0][j]),

    // Data Shift Register
    .write_shift_reg(write_shift_reg),
    .read_shift_reg(read_shift_reg),
    .shift_in(data_vector_in[(BIT_WIDTH*(j+1))-1:BIT_WIDTH*j]),
    .shift_out(shift_reg[0][j]),

    // Type Shift Register
    .shift_t_in(type_vector_in[(TYPE_BIT_WIDTH*(j+1))-1:TYPE_BIT_WIDTH*j]),
    .shift_t_out(shift_reg_t[0][j])

  );
end // y loop

endgenerate

/******************************************************************************
* Corners
******************************************************************************/

// Top left processing element (0, HEIGHT-1)
processing_element #(
  .BIT_WIDTH(BIT_WIDTH),
  .SPACE_START(32'h00000000),
  .NUM_REGS(NUM_REGS),
  .NUM_BITS(NUM_BITS),
  .TYPE_BIT_WIDTH(TYPE_BIT_WIDTH),
  .RULE_DELAY(RULE_DELAY),
  .START_TYPE(START_TYPE)
)
top_left (
  .clock(clock),
  .reset_n(reset_n),

  .read_rq_from_neighbor( { read_rq[WIDTH-1][HEIGHT-1][2], read_rq[1][HEIGHT-1][3],
                          read_rq[0][HEIGHT-2][0], read_rq[0][0][1]} ),
  .empty_from_neighbor( { empty[WIDTH-1][HEIGHT-1][2], empty[1][HEIGHT-1][3],
                          empty[0][HEIGHT-2][0], empty[0][0][1] } ),
  .north_in(space_fifo_s[0][0]),
  .south_in(space_fifo_n[0][HEIGHT-2]),
  .east_in(space_fifo_w[1][HEIGHT-1]),
  .west_in(space_fifo_e[WIDTH-1][HEIGHT-1]),

  .read_rq_to_neighbor(read_rq[0][HEIGHT-1]),
  .empty_to_neighbor(empty[0][HEIGHT-1]),

  .north_out(space_fifo_n[0][HEIGHT-1]),
  .south_out(space_fifo_s[0][HEIGHT-1]),
  .east_out(space_fifo_e[0][HEIGHT-1]),
  .west_out(space_fifo_w[0][HEIGHT-1]),

  // Signals to move Agent to adjacent PE
  .cell_state_n(cell_state[0][0]),              // north
  .cell_state_s(cell_state[0][HEIGHT-2]),       // south
  .cell_state_e(cell_state[1][HEIGHT-1]),       // east
  .cell_state_w(cell_state[WIDTH-1][HEIGHT-1]), // west
  .cell_state_out(cell_state[0][HEIGHT-1]),     // output

  // Sender Side
  .type0_n_in(type0[0][0]),
  .type0_s_in(type0[0][HEIGHT-2]),
  .type0_e_in(type0[1][HEIGHT-1]),
  .type0_w_in(type0[WIDTH-1][HEIGHT-1]),

  .move_rq_n_out(move_rq_n[0][HEIGHT-1]),
  .move_rq_s_out(move_rq_s[0][HEIGHT-1]),
  .move_rq_e_out(move_rq_e[0][HEIGHT-1]),
  .move_rq_w_out(move_rq_w[0][HEIGHT-1]),

  .type_update_n_in(update_s[0][0]),              // The south wire of the northern PE
  .type_update_s_in(update_n[0][HEIGHT-2]),       // The north wire of the southern PE
  .type_update_e_in(update_w[1][HEIGHT-1]),       // The west wire of the eastern PE
  .type_update_w_in(update_e[WIDTH-1][HEIGHT-1]), // The east wire of the western PE

  // Reciever Side
  .type0_out(type0[0][HEIGHT-1]),

  .move_rq_n_in(move_rq_s[0][0]),              // The south wire of the northern PE
  .move_rq_s_in(move_rq_n[0][HEIGHT-2]),       // The north wire of the southern PE
  .move_rq_e_in(move_rq_w[1][HEIGHT-1]),       // The west wire of the eastern PE
  .move_rq_w_in(move_rq_e[WIDTH-1][HEIGHT-1]), // The east wire of the western PE

  .type_update_n_out(update_n[0][HEIGHT-1]),
  .type_update_s_out(update_s[0][HEIGHT-1]),
  .type_update_e_out(update_e[0][HEIGHT-1]),
  .type_update_w_out(update_w[0][HEIGHT-1]),

  // Data Shift Register
  .write_shift_reg(write_shift_reg),
  .read_shift_reg(read_shift_reg),
  .shift_in(data_vector_in[(BIT_WIDTH*HEIGHT)-1:(BIT_WIDTH*(HEIGHT-1))]),
  .shift_out(shift_reg[0][HEIGHT-1]),

  // Type Shift Register
  .shift_t_in(type_vector_in[(TYPE_BIT_WIDTH*HEIGHT)-1:TYPE_BIT_WIDTH*(HEIGHT-1)]),
  .shift_t_out(shift_reg_t[0][HEIGHT-1])

);

// Top right processing element (WIDTH-1, HEIGHT-1)
processing_element #(
  .BIT_WIDTH(BIT_WIDTH),
  .SPACE_START(32'h00000000),
  .NUM_REGS(NUM_REGS),
  .NUM_BITS(NUM_BITS),
  .TYPE_BIT_WIDTH(TYPE_BIT_WIDTH),
  .RULE_DELAY(RULE_DELAY),
  .START_TYPE(START_TYPE)
)
top_right (
  .clock(clock),
  .reset_n(reset_n),

  .read_rq_from_neighbor( {read_rq[WIDTH-2][HEIGHT-1][2], read_rq[0][HEIGHT-1][3],
                         read_rq[WIDTH-1][HEIGHT-2][0], read_rq[WIDTH-1][0][1]} ),
  .empty_from_neighbor( {empty[WIDTH-2][HEIGHT-1][2], empty[0][HEIGHT-1][3],
                         empty[WIDTH-1][HEIGHT-2][0], empty[WIDTH-1][0][1]} ),

  .north_in(space_fifo_s[WIDTH-1][0]),
  .south_in(space_fifo_n[WIDTH-1][HEIGHT-2]),
  .east_in(space_fifo_w[0][HEIGHT-1]),
  .west_in(space_fifo_e[WIDTH-2][HEIGHT-1]),

  .read_rq_to_neighbor(read_rq[WIDTH-1][HEIGHT-1]),
  .empty_to_neighbor(empty[WIDTH-1][HEIGHT-1]),

  .north_out(space_fifo_n[WIDTH-1][HEIGHT-1]),
  .south_out(space_fifo_s[WIDTH-1][HEIGHT-1]),
  .east_out(space_fifo_e[WIDTH-1][HEIGHT-1]),
  .west_out(space_fifo_w[WIDTH-1][HEIGHT-1]),

  // Signals to move Agent to adjacent PE
  .cell_state_n(cell_state[WIDTH-1][0]),          // north
  .cell_state_s(cell_state[WIDTH-1][HEIGHT-2]),   // south
  .cell_state_e(cell_state[0][HEIGHT-1]),         // east
  .cell_state_w(cell_state[WIDTH-2][HEIGHT-1]),   // west
  .cell_state_out(cell_state[WIDTH-1][HEIGHT-1]), // output

  // Sender Side
  .type0_n_in(type0[WIDTH-1][0]),
  .type0_s_in(type0[WIDTH-1][HEIGHT-2]),
  .type0_e_in(type0[0][HEIGHT-1]),
  .type0_w_in(type0[WIDTH-2][HEIGHT-1]),

  .move_rq_n_out(move_rq_n[WIDTH-1][HEIGHT-1]),
  .move_rq_s_out(move_rq_s[WIDTH-1][HEIGHT-1]),
  .move_rq_e_out(move_rq_e[WIDTH-1][HEIGHT-1]),
  .move_rq_w_out(move_rq_w[WIDTH-1][HEIGHT-1]),

  .type_update_n_in(update_s[WIDTH-1][0]),        // The south wire of the northern PE
  .type_update_s_in(update_n[WIDTH-1][HEIGHT-2]), // The north wire of the southern PE
  .type_update_e_in(update_w[0][HEIGHT-1]),       // The west wire of the eastern PE
  .type_update_w_in(update_e[WIDTH-2][HEIGHT-1]), // The east wire of the western PE

  // Reciever Side
  .type0_out(type0[WIDTH-1][HEIGHT-1]),

  .move_rq_n_in(move_rq_s[WIDTH-1][0]),        // The south wire of the northern PE
  .move_rq_s_in(move_rq_n[WIDTH-1][HEIGHT-2]), // The north wire of the southern PE
  .move_rq_e_in(move_rq_w[0][HEIGHT-1]),       // The west wire of the eastern PE
  .move_rq_w_in(move_rq_e[WIDTH-2][HEIGHT-1]), // The east wire of the western PE

  .type_update_n_out(update_n[WIDTH-1][HEIGHT-1]),
  .type_update_s_out(update_s[WIDTH-1][HEIGHT-1]),
  .type_update_e_out(update_e[WIDTH-1][HEIGHT-1]),
  .type_update_w_out(update_w[WIDTH-1][HEIGHT-1]),

  // Data Shift Register
  .write_shift_reg(write_shift_reg),
  .read_shift_reg(read_shift_reg),
  .shift_in(shift_reg[WIDTH-2][HEIGHT-1]),
  .shift_out(data_vector_out[(BIT_WIDTH*HEIGHT)-1:BIT_WIDTH*(HEIGHT-1)]),

  // Type Shift Register
  .shift_t_in(shift_reg_t[WIDTH-2][HEIGHT-1]),
  .shift_t_out(type_vector_out[(TYPE_BIT_WIDTH*HEIGHT)-1:TYPE_BIT_WIDTH*(HEIGHT-1)])

);

// Bottom left processing element (0,0)
processing_element #(
  .BIT_WIDTH(BIT_WIDTH),
  .SPACE_START(32'h00000000),
  .NUM_REGS(NUM_REGS),
  .NUM_BITS(NUM_BITS),
  .TYPE_BIT_WIDTH(TYPE_BIT_WIDTH),
  .RULE_DELAY(RULE_DELAY),
  .START_TYPE(START_TYPE)
)
bottom_left (
  .clock(clock),
  .reset_n(reset_n),

  .read_rq_from_neighbor( { read_rq[WIDTH-1][0][2], read_rq[1][0][3],
                          read_rq[0][HEIGHT-1][0], read_rq[0][1][1] } ),
  .empty_from_neighbor( { empty[WIDTH-1][0][2], empty[1][0][3],
                          empty[0][HEIGHT-1][0], empty[0][1][1] } ),

  .north_in(space_fifo_s[0][1]),
  .south_in(space_fifo_n[0][HEIGHT-1]),
  .east_in(space_fifo_w[1][0]),
  .west_in(space_fifo_e[WIDTH-1][0]),

  .read_rq_to_neighbor(read_rq[0][0]),
  .empty_to_neighbor(empty[0][0]),

  .north_out(space_fifo_n[0][0]),
  .south_out(space_fifo_s[0][0]),
  .east_out(space_fifo_e[0][0]),
  .west_out(space_fifo_w[0][0]),

  // Signals to move Agent to adjacent PE
  .cell_state_n(cell_state[0][1]),        // north
  .cell_state_s(cell_state[0][HEIGHT-1]), // south
  .cell_state_e(cell_state[1][0]),        // east
  .cell_state_w(cell_state[WIDTH-1][0]),  // west
  .cell_state_out(cell_state[0][0]),      // output

  // Sender Side
  .type0_n_in(type0[0][1]),
  .type0_s_in(type0[0][HEIGHT-1]),
  .type0_e_in(type0[1][0]),
  .type0_w_in(type0[WIDTH-1][0]),

  .move_rq_n_out(move_rq_n[0][0]),
  .move_rq_s_out(move_rq_s[0][0]),
  .move_rq_e_out(move_rq_e[0][0]),
  .move_rq_w_out(move_rq_w[0][0]),

  .type_update_n_in(update_s[0][1]),        // The south wire of the northern PE
  .type_update_s_in(update_n[0][HEIGHT-1]), // The north wire of the southern PE
  .type_update_e_in(update_w[1][0]),        // The west wire of the eastern PE
  .type_update_w_in(update_e[WIDTH-1][0]),  // The east wire of the western PE

  // Reciever Side
  .type0_out(type0[0][0]),

  .move_rq_n_in(move_rq_s[0][1]),        // The south wire of the northern PE
  .move_rq_s_in(move_rq_n[0][HEIGHT-1]), // The north wire of the southern PE
  .move_rq_e_in(move_rq_w[1][0]),        // The west wire of the eastern PE
  .move_rq_w_in(move_rq_e[WIDTH-1][0]),  // The east wire of the western PE

  .type_update_n_out(update_n[0][0]),
  .type_update_s_out(update_s[0][0]),
  .type_update_e_out(update_e[0][0]),
  .type_update_w_out(update_w[0][0]),

  // Data Shift Register
  .write_shift_reg(write_shift_reg),
  .read_shift_reg(read_shift_reg),
  .shift_in(data_vector_in[BIT_WIDTH-1:0]),
  .shift_out(shift_reg[0][0]),

  // Type Shift Register
  .shift_t_in(type_vector_in[TYPE_BIT_WIDTH-1:0]),
  .shift_t_out(shift_reg_t[0][0])

);

// Bottom right processing element (Width-1, 0)
processing_element #(
  .BIT_WIDTH(BIT_WIDTH),
  .SPACE_START(32'h00000000),
  .NUM_REGS(NUM_REGS),
  .NUM_BITS(NUM_BITS),
  .TYPE_BIT_WIDTH(TYPE_BIT_WIDTH),
  .RULE_DELAY(RULE_DELAY),
  .START_TYPE(START_TYPE)
)
bottom_right (
  .clock(clock),
  .reset_n(reset_n),
  .read_rq_from_neighbor( { read_rq[WIDTH-2][0][2], read_rq[0][0][3],
                          read_rq[WIDTH-1][HEIGHT-1][0], read_rq[WIDTH-1][1][1] } ),
  .empty_from_neighbor( { empty[WIDTH-2][0][2], empty[0][0][3],
                          empty[WIDTH-1][HEIGHT-1][0], empty[WIDTH-1][1][1] } ),

  .north_in(space_fifo_s[WIDTH-1][1]),
  .south_in(space_fifo_n[WIDTH-1][HEIGHT-1]),
  .east_in(space_fifo_w[0][0]),
  .west_in(space_fifo_e[WIDTH-2][0]),

  .read_rq_to_neighbor(read_rq[WIDTH-1][0]),
  .empty_to_neighbor(empty[WIDTH-1][0]),

  .north_out(space_fifo_n[WIDTH-1][0]),
  .south_out(space_fifo_s[WIDTH-1][0]),
  .east_out(space_fifo_e[WIDTH-1][0]),
  .west_out(space_fifo_w[WIDTH-1][0]),

  // Signals to move Agent to adjacent PE
  .cell_state_n(cell_state[WIDTH-1][1]),        // north
  .cell_state_s(cell_state[WIDTH-1][HEIGHT-1]), // south
  .cell_state_e(cell_state[0][0]),              // east
  .cell_state_w(cell_state[WIDTH-2][0]),        // west
  .cell_state_out(cell_state[WIDTH-1][0]),      // output

  // Sender Side
  .type0_n_in(type0[WIDTH-1][1]),
  .type0_s_in(type0[WIDTH-1][HEIGHT-1]),
  .type0_e_in(type0[0][0]),
  .type0_w_in(type0[WIDTH-2][0]),

  .move_rq_n_out(move_rq_n[WIDTH-1][0]),
  .move_rq_s_out(move_rq_s[WIDTH-1][0]),
  .move_rq_e_out(move_rq_e[WIDTH-1][0]),
  .move_rq_w_out(move_rq_w[WIDTH-1][0]),

  .type_update_n_in(update_s[WIDTH-1][1]),        // The south wire of the northern PE
  .type_update_s_in(update_n[WIDTH-1][HEIGHT-1]), // The north wire of the southern PE
  .type_update_e_in(update_w[0][0]),              // The west wire of the eastern PE
  .type_update_w_in(update_e[WIDTH-2][0]),        // The east wire of the western PE

  // Reciever Side
  .type0_out(type0[WIDTH-1][0]),

  .move_rq_n_in(move_rq_s[WIDTH-1][1]),        // The south wire of the northern PE
  .move_rq_s_in(move_rq_n[WIDTH-1][HEIGHT-1]), // The north wire of the southern PE
  .move_rq_e_in(move_rq_w[0][0]),              // The west wire of the eastern PE
  .move_rq_w_in(move_rq_e[WIDTH-2][0]),        // The east wire of the western PE

  .type_update_n_out(update_n[WIDTH-1][0]),
  .type_update_s_out(update_s[WIDTH-1][0]),
  .type_update_e_out(update_e[WIDTH-1][0]),
  .type_update_w_out(update_w[WIDTH-1][0]),

  // Data Shift Register
  .write_shift_reg(write_shift_reg),
  .read_shift_reg(read_shift_reg),
  .shift_in(shift_reg[WIDTH-2][0]),
  .shift_out(data_vector_out[BIT_WIDTH-1:0]),

  // Type Shift Register
  .shift_t_in(shift_reg_t[WIDTH-2][0]),
  .shift_t_out(type_vector_out[TYPE_BIT_WIDTH-1:0])

);

endmodule
