/* This test bench simulates a 3x3 PE architecture.
 * Inputs are loaded into the shift register and the simulation state is
 * output with the shift register and printed with $display commands.
 */

`timescale 1ps / 1ps
module torus_tb();

parameter BIT_WIDTH = 32; // The bit width of the data
parameter NUM_REGS = 4;
parameter NUM_BITS = NUM_REGS * BIT_WIDTH;
parameter TYPE_BIT_WIDTH = 1;
parameter RULE_DELAY = 1;
parameter HEIGHT = 3;
parameter WIDTH = 3;

parameter CLK = 10;
parameter HALF_CLK = CLK/2;
parameter TICK = (1 + NUM_REGS) * CLK;

reg clock;
reg reset_n;

reg write_shift_reg;
reg read_shift_reg;
reg [(HEIGHT*BIT_WIDTH)-1:0] data_vector_in;
reg [(HEIGHT*TYPE_BIT_WIDTH)-1:0] type_vector_in;

wire [(HEIGHT*BIT_WIDTH)-1:0] data_vector_out;
wire [(HEIGHT*TYPE_BIT_WIDTH)-1:0] type_vector_out;

wire [BIT_WIDTH-1:0] chem;
wire [BIT_WIDTH-1:0] chem_sr;
wire [NUM_BITS-1:0] r_cell;
wire [TYPE_BIT_WIDTH-1:0] t;

assign chem = dut.top_right.space_pe.core.value;
assign chem_sr = dut.top_right.space_pe.core.shift_out;
assign r_cell = dut.top_right.cell_pe.control.cell_state_out;
assign t = dut.top_right.cell_pe.control.type_out;

torus #(
  .BIT_WIDTH(BIT_WIDTH),           // The bit width of the data
  .NUM_REGS(NUM_REGS),
  .NUM_BITS(NUM_BITS),
  .TYPE_BIT_WIDTH(TYPE_BIT_WIDTH),
  .RULE_DELAY(RULE_DELAY),
  .WIDTH(WIDTH),
  .HEIGHT(HEIGHT)
) dut (
  .clock(clock),
  .reset_n(reset_n),

  .write_shift_reg(write_shift_reg),
  .read_shift_reg(read_shift_reg),

  .data_vector_in(data_vector_in),
  .data_vector_out(data_vector_out),

  .type_vector_in(type_vector_in),
  .type_vector_out(type_vector_out)

);


always begin
  #HALF_CLK clock <= ~clock;
end


initial begin
  $display("############################################");
  $display("                 Torus TB                #");
  $display("############################################");

  clock = 1'b1;
  reset_n = 1'b1;

  write_shift_reg = 1'b0;
  read_shift_reg = 1'b0;
  data_vector_in = {HEIGHT*BIT_WIDTH{1'b0}};
  type_vector_in = {HEIGHT*TYPE_BIT_WIDTH{1'b0}};

  #10 reset_n = 1'b0;
  #10 reset_n = 1'b1;

  //data_vector_in = { 32'haaaaaaaa, 32'haaaaaaaa, 32'haaaaaaaa};
  //type_vector_in = { 1'b1, 1'b0, 1'b1};
  #10
  // Right column chem
  data_vector_in = { 32'h00000000, 32'h00000000, 32'h00000000 };
  type_vector_in = { 1'b0, 1'b0, 1'b0};
  #10
  // Right column cells
  data_vector_in = { 32'h0000000f, 32'h00000000, 32'h00000000 };
  type_vector_in = { 1'b0, 1'b0, 1'b0};
  #40
  // Middle column chem
  data_vector_in = { 32'h00000000, 32'h000f0000, 32'h00000000 };
  type_vector_in = { 1'b0, 1'b1, 1'b0};
  #10
  // Middle column cells
  data_vector_in = { 32'h00000000, 32'h00000000, 32'h00000000 };
  type_vector_in = { 1'b0, 1'b1, 1'b0};
  #40
  // Left column chem
  data_vector_in = { 32'h00000000, 32'h00000000, 32'h00000000 };
  type_vector_in = { 1'b0, 1'b0, 1'b0};
  #10
  // Left column cells
  data_vector_in = {(HEIGHT*BIT_WIDTH){1'b0}};
  type_vector_in = { 1'b0, 1'b0, 1'b0};
  #40
  read_shift_reg = 1'b1;

  $display("Loading Sr at time %d\n",$time);
  $display("Value of rotated chem field at time %d\n",$time);
  $display("TR\t\t\tBR\tTypes\n");
  $display("%h, %h, %h\t\t\t%h, %h, %h\n", data_vector_out[95:64], data_vector_out[63:32],
    data_vector_out[31:0], type_vector_out[2], type_vector_out[1], type_vector_out[0]);

  #10
  read_shift_reg = 1'b0;

  /* No need to write sr, it is already writen by data_vector_in */

  #(TICK-10)
  $display("%h, %h, %h\t\t\t%h, %h, %h\n", data_vector_out[95:64], data_vector_out[63:32],
    data_vector_out[31:0], type_vector_out[2], type_vector_out[1], type_vector_out[0]);
  #TICK
  $display("%h, %h, %h\t\t\t%h, %h, %h\n", data_vector_out[95:64], data_vector_out[63:32],
    data_vector_out[31:0], type_vector_out[2], type_vector_out[1], type_vector_out[0]);
  $display("TL\t\t\tBL\n\n");
  #TICK

  #10 write_shift_reg = 1'b1;
  #10 write_shift_reg = 1'b0;
  $display("Loading Sr at time %d\n",$time);
  $display("Value of rotated chem field at time %d\n",$time);
  $display("TR\t\t\tBR\tTypes\n");
  $display("%h, %h, %h\t\t\t%h, %h, %h\n", data_vector_out[95:64], data_vector_out[63:32],
    data_vector_out[31:0], type_vector_out[2], type_vector_out[1], type_vector_out[0]);
  #TICK
  $display("%h, %h, %h\t\t\t%h, %h, %h\n", data_vector_out[95:64], data_vector_out[63:32],
    data_vector_out[31:0], type_vector_out[2], type_vector_out[1], type_vector_out[0]);
  #TICK
  $display("%h, %h, %h\t\t\t%h, %h, %h\n", data_vector_out[95:64], data_vector_out[63:32],
    data_vector_out[31:0], type_vector_out[2], type_vector_out[1], type_vector_out[0]);
  $display("TL\t\t\tBL\n\n");
  #TICK

  #10 write_shift_reg = 1'b1;
  #10 write_shift_reg = 1'b0;
  $display("Loading Sr at time %d\n",$time);
  $display("Value of rotated chem field at time %d\n",$time);
  $display("TR\t\t\tBR\tTypes\n");
  $display("%h, %h, %h\t\t\t%h, %h, %h\n", data_vector_out[95:64], data_vector_out[63:32],
    data_vector_out[31:0], type_vector_out[2], type_vector_out[1], type_vector_out[0]);
  #TICK
  $display("%h, %h, %h\t\t\t%h, %h, %h\n", data_vector_out[95:64], data_vector_out[63:32],
    data_vector_out[31:0], type_vector_out[2], type_vector_out[1], type_vector_out[0]);
  #TICK
  $display("%h, %h, %h\t\t\t%h, %h, %h\n", data_vector_out[95:64], data_vector_out[63:32],
    data_vector_out[31:0], type_vector_out[2], type_vector_out[1], type_vector_out[0]);
  $display("TL\t\t\tBL\n\n");
  #TICK


  #10 write_shift_reg = 1'b1;
  #10 write_shift_reg = 1'b0;
  $display("Loading Sr at time %d\n",$time);
  $display("Value of rotated chem field at time %d\n",$time);
  $display("TR\t\t\tBR\tTypes\n");
  $display("%h, %h, %h\t\t\t%h, %h, %h\n", data_vector_out[95:64], data_vector_out[63:32],
    data_vector_out[31:0], type_vector_out[2], type_vector_out[1], type_vector_out[0]);
  #TICK
  $display("%h, %h, %h\t\t\t%h, %h, %h\n", data_vector_out[95:64], data_vector_out[63:32],
    data_vector_out[31:0], type_vector_out[2], type_vector_out[1], type_vector_out[0]);
  #TICK
  $display("%h, %h, %h\t\t\t%h, %h, %h\n", data_vector_out[95:64], data_vector_out[63:32],
    data_vector_out[31:0], type_vector_out[2], type_vector_out[1], type_vector_out[0]);
  $display("TL\t\t\tBL\n\n");
  #TICK

  #1000
  $display("\n");
  $stop;


end

endmodule


